import rclpy
from rclpy.node import Node
from behavior_interface.msg import Command, BehaviorStatus
import time

class CommandNode(Node):
    def __init__(self):
        super().__init__('command_node')

        self.behaviors = ['Hover', 'MoveLeft', 'MoveForward', 'MoveRight', 'TakeOff', 'StopAndLand', 'MoveUp', 'MoveDown', 'MoveForwardVp', 'AlignCorridor', 'CenterCorridor','TurnRight', 'TurnLeft']

        self.commands = {
            'Stop': [(0, 'Hover')],
            'Dance': [(0, 'MoveLeft'), (0, 'MoveForward'), (2.5, 'MoveRight'), (3.2, 'Hover')],
            'TakeOff': [(0, 'TakeOff')],
            'StopAndLand': [(0, 'StopAndLand')],
            'MoveLeft': [(0, 'MoveLeft')],
            'MoveRight': [(0, 'MoveRight')],
            'Hover': [(0, 'Hover')],
            'MoveForward': [(0, 'MoveForward')],
            'MoveUp': [(0, 'MoveUp')],
            'MoveDown': [(0, 'MoveDown')],
            'TurnLeft': [(0, 'TurnLeft')],
            'TurnRight': [(0, 'TurnRight')],
            'GoAhead': [(0, 'MoveForwardVp'), (0, 'AlignCorridor'), (0, 'CenterCorridor')]  
        }

        self.command_subscription = self.create_subscription(
            Command,
            '/command',
            self.command_callback,
            10
        )
        self.behavior_publisher = self.create_publisher(BehaviorStatus, 'behavior', 10)

        self.event_queue = []
        self.timer = self.create_timer(0.05, self._on_time)
        self.active_behavior = None  

    def command_callback(self, msg):
        command_name = msg.command
        if command_name not in self.commands:
            self.get_logger().error(f"Unknown command: {command_name}")
            return

        self.deactivate_all_behaviors()

        current_time = time.time()
        for delay, behavior_name in self.commands[command_name]:
            event_time = current_time + delay
            self.event_queue.append((event_time, behavior_name))

        self.event_queue.sort()

    def deactivate_all_behaviors(self):
        for behavior in self.behaviors:
            self.publish_behavior(behavior, False)
        self.active_behavior = None  
        self.get_logger().info("All behaviors deactivated")

    def _on_time(self):
        current_time = time.time()
        
        while self.event_queue and self.event_queue[0][0] <= current_time:
            _, behavior_name = self.event_queue.pop(0)
            self.publish_behavior(behavior_name, True)
            self.active_behavior = behavior_name  

        if not self.event_queue and self.active_behavior != 'Hover':
            self.publish_behavior('Hover', True)
            self.active_behavior = 'Hover'  

    def publish_behavior(self, behavior_name, activate):
        status_msg = BehaviorStatus()
        status_msg.name = behavior_name
        status_msg.status = activate
        self.behavior_publisher.publish(status_msg)
        self.get_logger().info(f"Behavior {behavior_name} {'activated' if activate else 'deactivated'}")

def main(args=None):
    rclpy.init(args=args)
    command_node = CommandNode()
    rclpy.spin(command_node)
    command_node.destroy_node()
    rclpy.shutdown()

