#!/usr/bin/env python

import rclpy
from rclpy.node import Node
from behavior_interface.msg import Command
from sensor_msgs.msg import Joy

BUTTON_A = 0
BUTTON_B = 1
BUTTON_X = 2
BUTTON_Y = 3
BUTTON_LB = 4  # Left shoulder button for MoveUp
BUTTON_RB = 5  # Right shoulder button for MoveDown
BUTTON_RT = 5
AXIS_LEFT_VERTICAL = 1
AXIS_RIGHT_HORIZONTAL = 3  # Right joystick horizontal axis for rotation control

class TeleopNode(Node):
    def __init__(self, node_name='teleop'):
        super().__init__(node_name)
        
        self.start_command = 'TakeOff'
        self.stop_command = 'StopAndLand'
        self.last_command = None
        self.deadman_pressed = False
        
        # Publisher for behavior commands
        self.command_publisher = self.create_publisher(Command, '/command', 10)
        # Joystick subscriber
        self.joy_subscription = self.create_subscription(Joy, '/joy', self.joy_callback, 10)
        # Timer to repeat the last command
        self.command_timer = self.create_timer(0.1, self.repeat_last_command)
        
        # Tolerance for the deadman button
        self.axis_tolerance = 0.75 

    def joy_callback(self, msg):
        # Check the status of the RT button for takeoff/landing control
        rt_pressed = msg.axes[BUTTON_RT] < -1.0 + self.axis_tolerance  # Status of button press
        
        if rt_pressed and not self.deadman_pressed:
            # Trigger takeoff if RT button is just pressed
            self.deadman_pressed = True
            self.send_command(self.start_command)
        elif not rt_pressed and self.deadman_pressed:
            # Trigger landing if RT button is released
            self.deadman_pressed = False
            self.send_command(self.stop_command)

        # Process other joystick controls when RT is pressed
        if rt_pressed:
            self.process_joystick_commands(msg)

    def send_command(self, command_str):
        # Publish command only if it changes
        if command_str != self.last_command or command_str in [self.start_command, self.stop_command]:
            command_msg = Command()
            command_msg.command = command_str
            self.command_publisher.publish(command_msg)
            self.get_logger().info(f"Published command: {command_str}")
            self.last_command = command_str

    def repeat_last_command(self):
        # Repeat the last command if the deadman button is pressed
        if self.last_command and self.deadman_pressed:
            self.send_command(self.last_command)

    def process_joystick_commands(self, msg):
        # Mapping from buttons to behaviors
        command_mappings = {
            BUTTON_A: 'Hover',
            BUTTON_B: 'MoveRight',
            BUTTON_X: 'MoveLeft',
            BUTTON_Y: 'MoveForward',
            BUTTON_LB: 'MoveUp',   # Left shoulder button for MoveUp
            BUTTON_RB: 'MoveDown'  # Right shoulder button for MoveDown
        }

        # Process commands based on button press
        for button_index, command in command_mappings.items():
            if msg.buttons[button_index] == 1:
                self.send_command(command)
                return  # Process only one command

        # Process vertical axis of left joystick for forward movement with a deadzone
        deadzone = 0.1
        if abs(msg.axes[AXIS_LEFT_VERTICAL]) > deadzone:
            vertical_command = 'MoveForward' if msg.axes[AXIS_LEFT_VERTICAL] > 0 else 'Hover'
            self.send_command(vertical_command)
        
        # Process horizontal axis of right joystick for rotation
        if msg.axes[AXIS_RIGHT_HORIZONTAL] > deadzone:
            self.send_command('TurnRight')
        elif msg.axes[AXIS_RIGHT_HORIZONTAL] < -deadzone:
            self.send_command('TurnLeft')

def main(args=None):
    rclpy.init(args=args)
    teleop = TeleopNode()
    rclpy.spin(teleop)
    teleop.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
