#!/usr/bin/env python

import rclpy
from rclpy.node import Node
from geometry_msgs.msg import Twist
from std_msgs.msg import Empty, Float32, Bool
from sensor_msgs.msg import Joy

# Joystick button and axis mappings
BUTTON_RT = 5                # Mapping for the RT button
AXIS_LEFT_HORIZONTAL = 0      # Left joystick horizontal axis (left/right)
AXIS_LEFT_VERTICAL = 1        # Left joystick vertical axis (up/down)
AXIS_RIGHT_HORIZONTAL = 3     # Right joystick horizontal axis (left/right)
AXIS_RIGHT_VERTICAL = 4       # Right joystick vertical axis (up/down)

class JoyTeleop(Node):
    def __init__(self):
        super().__init__('joy_teleop')

        # Create publishers
        self.linear_x_pub = self.create_publisher(Float32, "/linear_x", 10)
        self.linear_y_pub = self.create_publisher(Float32, "/linear_y", 10)
        self.linear_z_pub = self.create_publisher(Float32, "/linear_z", 10)
        self.angular_z_pub = self.create_publisher(Float32, "/angular_z", 10)
        self.hover_pub = self.create_publisher(Bool, "/hover", 10)
        self.takeoff_pub = self.create_publisher(Empty, '/bebop/takeoff', 10)
        self.land_pub = self.create_publisher(Empty, '/bebop/land', 10)

        # Create subscriber
        # self.joy_sub = self.create_subscription(Joy, '/joy', self.joy_callback, 10)

        # Initialize state
        self.is_flying = False
        self.axis_tolerance = 0.75  # Threshold for detecting button press

    def joy_callback(self, msg):
        # Check the state of the RT button
        if msg.axes[BUTTON_RT] < -1.0 + self.axis_tolerance:
            if not self.is_flying:
                # Takeoff control
                self.takeoff_pub.publish(Empty())
                self.is_flying = True
                self.get_logger().info("Takeoff triggered")
            else:
                # RT button is held down, publish velocity commands
                hover_msg = Bool()
                hover_msg.data = False  # Not in hover mode
                self.hover_pub.publish(hover_msg)

                # Publish linear and angular velocities
                linear_x = Float32()
                linear_y = Float32()
                linear_z = Float32()
                angular_z = Float32()

                # Set linear and angular velocity values
                linear_x.data = msg.axes[AXIS_LEFT_VERTICAL] * 0.5
                linear_y.data = msg.axes[AXIS_LEFT_HORIZONTAL] * 0.5
                linear_z.data = msg.axes[AXIS_RIGHT_VERTICAL] * 0.5
                angular_z.data = msg.axes[AXIS_RIGHT_HORIZONTAL] * 0.5

                # Publish velocity commands
                self.linear_x_pub.publish(linear_x)
                self.linear_y_pub.publish(linear_y)
                self.linear_z_pub.publish(linear_z)
                self.angular_z_pub.publish(angular_z)

        else:
            # If RT is released, trigger landing
            if self.is_flying:
                self.land_pub.publish(Empty())
                self.is_flying = False
                self.get_logger().info("Landing triggered")

def main(args=None):
    rclpy.init(args=args)
    joy_teleop = JoyTeleop()
    rclpy.spin(joy_teleop)
    joy_teleop.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
