import rclpy
import math
from rclpy.node import Node
from geometry_msgs.msg import Twist, Point
from std_msgs.msg import Bool, Float32
from nav_msgs.msg import Odometry
from sensor_msgs.msg import Joy
from tf2_ros import Buffer, TransformListener
from tf_transformations import euler_from_quaternion
import time

# Define button mappings for the remote controller
AXIS_CROSS_HORIZONTAL = 6  # D-pad horizontal direction

class Slide(Node):
    def __init__(self):
        super().__init__('slide_behavior')
        
        # Publishers and Subscribers
        self.status_subscriber = self.create_subscription(Bool, 'slide_status', self.on_status_on, 10)
        self.odom_subscriber = self.create_subscription(Odometry, '/odom', self.odom_callback, 10)
        self.joy_subscriber = self.create_subscription(Joy, '/joy', self.joy_callback, 10)

        self.linear_x_pub = self.create_publisher(Float32, '/linear_x', 10)
        self.linear_y_pub = self.create_publisher(Float32, '/linear_y', 10)
        self.angular_z_pub = self.create_publisher(Float32, '/angular_z', 10)

        # TF2 Buffer and Listener to capture `base_link` position
        self.tf_buffer = Buffer()
        self.tf_listener = TransformListener(self.tf_buffer, self)

        # Timer for controlling the open-loop behavior
        self.timer = None
        self.is_active = False

        # Store initial position and quaternion for alignment
        self.initial_position = None
        self.initial_quaternion = None
        self.current_position = None
        self.current_orientation = None

    def on_status_on(self, msg):
        # Trigger slide behavior if not already active
        if msg.data and not self.is_active:
            self.get_logger().info("Slide behavior activated.")
            self.is_active = True
            self.start_slide()
    
    def start_slide(self):
        # Record initial position of `base_link`
        initial_transform = self.get_base_link_transform()
        if initial_transform:
            self.initial_position = initial_transform['position']
            self.initial_quaternion = initial_transform['quaternion']
            self.get_logger().info(f"Initial position of base_link: {self.initial_position}")
            self.get_logger().info(f"Initial quaternion of base_link: {self.initial_quaternion}")
        else:
            self.get_logger().warn("Could not get initial position of base_link.")
            return

        # Step 1: Move forward along x and rotate around z for 1 second
        self.perform_maneuver(vel_x=0.5, ang_z=3.0, duration=1.0)

        # Step 2: Slide laterally along y
        duration = 0
        while duration <= 1:
            duration += 0.1
            vel_x, vel_y, theta = self.instruction_regain_main_lane(vel_slide=0.5)
            self.perform_maneuver(vel_y=vel_y, ang_z=-theta / math.pi, duration=0.1)

        # Deactivate after maneuver
        self.is_active = False
        self.get_logger().info("Slide behavior completed and deactivated.")

    def instruction_regain_main_lane(self, vel_slide):
        # Calculate alignment and regain movement
        drone_ref = self.get_base_link_transform()
        if not drone_ref:
            return 0.0, 0.0, 0.0
        
        initial_angle_ref = self.quaternion_to_euler(self.initial_quaternion)
        target_angle_ref = initial_angle_ref.copy()
        target_angle_ref['yaw'] -= math.pi / 2
        target_angle_ref['yaw'] = (target_angle_ref['yaw'] + math.pi) % (2 * math.pi) - math.pi  # Normalize

        drone_angle_ref = self.quaternion_to_euler(drone_ref['quaternion'])
        theta = drone_angle_ref['yaw'] - target_angle_ref['yaw']

        vel_x = vel_slide * (-math.sin(theta))
        vel_y = vel_slide * math.cos(theta)

        # Calculate distance to axis for alignment adjustment
        P0D = (
            drone_ref['position']['x'] - self.initial_position['x'],
            drone_ref['position']['y'] - self.initial_position['y']
        )
        vectorn = (math.sin(theta), -math.cos(theta))
        norme_normal_speed = max(0.1, (P0D[0] * vectorn[0] + P0D[1] * vectorn[1]))

        vel_x += vectorn[0] * norme_normal_speed
        vel_y += vectorn[1] * norme_normal_speed
        return vel_x, vel_y, theta

    def perform_maneuver(self, vel_x=0.0, vel_y=0.0, ang_z=0.0, duration=1.0):
        # Initialize Twist message for velocity commands
        twist_msg = Twist()
        twist_msg.linear.x = vel_x
        twist_msg.linear.y = vel_y
        twist_msg.angular.z = ang_z

        # Start maneuver by publishing twist message for given duration
        end_time = self.get_clock().now().to_msg().sec + duration
        while self.get_clock().now().to_msg().sec < end_time:
            self.linear_x_pub.publish(Float32(data=twist_msg.linear.x))
            self.linear_y_pub.publish(Float32(data=twist_msg.linear.y))
            self.angular_z_pub.publish(Float32(data=twist_msg.angular.z))
            time.sleep(0.1)  # Small delay to prevent excessive publishing

    def get_base_link_transform(self):
        try:
            # Get the transformation from `odom` to `base_link`
            trans = self.tf_buffer.lookup_transform('odom', 'base_link', rclpy.time.Time())
            translation = trans.transform.translation
            rotation = trans.transform.rotation

            # Return position and quaternion
            return {
                'position': {
                    'x': translation.x,
                    'y': translation.y,
                    'z': translation.z
                },
                'quaternion': {
                    'x': rotation.x,
                    'y': rotation.y,
                    'z': rotation.z,
                    'w': rotation.w
                }
            }

        except Exception as e:
            self.get_logger().warn(f"Transform not available: {e}")
            return None

    def quaternion_to_euler(self, quaternion):
        """Convert a quaternion into Euler angles (roll, pitch, yaw)."""
        q = (quaternion['x'], quaternion['y'], quaternion['z'], quaternion['w'])
        euler_angles = euler_from_quaternion(q)
        return {
            'roll': euler_angles[0],
            'pitch': euler_angles[1],
            'yaw': euler_angles[2]
        }

    def odom_callback(self, msg):
        # Callback to update current odometry data
        self.current_position = msg.pose.pose.position
        self.current_orientation = msg.pose.pose.orientation
        self.get_logger().info(f"Updated position: {self.current_position}")

    def joy_callback(self, msg):
        # Trigger slide behavior when D-pad left or right is pressed
        if msg.axes[AXIS_CROSS_HORIZONTAL] < -0.5 and not self.is_active:
            self.get_logger().info("Slide left behavior triggered by D-pad left.")
            self.is_active = True
            self.start_slide()
        elif msg.axes[AXIS_CROSS_HORIZONTAL] > 0.5 and not self.is_active:
            self.get_logger().info("Slide right behavior triggered by D-pad right.")
            self.is_active = True
            self.start_slide()

def main(args=None):
    rclpy.init(args=args)
    slide_behavior = Slide()
    rclpy.spin(slide_behavior)

    slide_behavior.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
