import rclpy
from rclpy.node import Node
from geometry_msgs.msg import Twist
from std_msgs.msg import Bool
import time
from sensor_msgs.msg import Joy

AXIS_CROSS_VERTICAL = 7 

class UTurnNode(Node):
    def __init__(self):
        super().__init__('u_turn_node')
        
        # Declare parameters with default values
        self.declare_parameter('u_turn.angular_velocity', 1.0)
        self.declare_parameter('u_turn.duration', 3.0)

        # get Parameters from launch file
        self.angular_velocity = self.get_parameter('u_turn.angular_velocity').value
        self.duration = self.get_parameter('u_turn.duration').value

        # Publisher to control robot velocity
        self.velocity_publisher = self.create_publisher(Twist, '/bebop/cmd_vel', 10)

        # # Subscriber to monitor status
        # self.status_subscriber = self.create_subscription(
        #     Bool,
        #     'u_turn_status',
        #     self.on_status_on,
        #     10
        # )
        
        # Subscriber to monitor D-pad input from the controller
        #self.joy_subscriber = self.create_subscription(
        #    Joy,
        #    '/joy',
        #    self.joy_callback,
        #    10
        #)

        # Timer for controlling the U-turn duration
        self.timer = None
        self.start_time = None
        self.is_turning = False

    # def on_status_on(self, msg):
    #     # Trigger the U-turn if status is active and not already turning
    #     if msg.data and not self.is_turning:
    #         self.get_logger().info("U-turn behavior triggered.")
    #         self.start_u_turn()

    def joy_callback(self, msg):
        # Trigger U-turn if D-pad up is pressed and not already turning
        if msg.axes[AXIS_CROSS_VERTICAL] > 0.5 and not self.is_turning:
            self.get_logger().info("U-turn behavior triggered by D-pad up.")
            self.start_u_turn()

    def start_u_turn(self):
        # Set flag to indicate the U-turn has started
        self.is_turning = True
        self.start_time = self.get_clock().now().to_msg().sec

        # Create a timer to execute the U-turn
        self.timer = self.create_timer(0.1, self.execute_u_turn)

    def execute_u_turn(self):
        # Calculate elapsed time
        current_time = self.get_clock().now().to_msg().sec
        elapsed_time = current_time - self.start_time

        # Check if we should continue the U-turn
        if elapsed_time < self.duration:
            u_turn_msg = Twist()
            u_turn_msg.angular.z = self.angular_velocity
            self.velocity_publisher.publish(u_turn_msg)
            self.get_logger().info(f"Executing U-turn... {elapsed_time:.2f}/{self.duration:.2f}s")
        else:
            # Stop the U-turn
            self.stop_u_turn()

    def stop_u_turn(self):
        # Publish zero velocity to stop rotation
        end_turn_msg = Twist()
        end_turn_msg.angular.z = 0.0
        self.velocity_publisher.publish(end_turn_msg)

        # Reset state
        self.is_turning = False
        if self.timer is not None:
            self.timer.cancel()
            self.timer = None

        self.get_logger().info("U-turn completed and behavior deactivated.")

def main(args=None):
    rclpy.init(args=args)
    u_turn_node = UTurnNode()
    rclpy.spin(u_turn_node)

    u_turn_node.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
