import numpy as np
import cv2
import sys
from sklearn.cluster import KMeans
from scipy.cluster.hierarchy import linkage, fcluster
import math

def from_lsd(lines_lsd):
   """
   lines_lsd : the weird output of cv line segment detector.
   returns : [[x1, y1, x2, y2, a, b, c, L],
              [...                       ],
              ...
             ]

             where 
               - [(x1, y1), (x2, y2)] is the segment (in image coordinates).
               - ax + by + c = 0 is the line equation
               - (a, b) is the normal vector of the line
               - |(a, b)| = 1, b >= 0
               - L is the length of the segment.
   """
   if lines_lsd is None :
       return np.array([]).reshape((0,8))

   nlines = lines_lsd.shape[0]
   segments = None

   if nlines == 0 :
       return np.array([]).reshape((0,8))

   if nlines == 1 :
       segments = lines_lsd.squeeze().reshape((1,4))
   else :
       segments = lines_lsd.squeeze()

   x1 = segments[:, 0]
   y1 = segments[:, 1]
   x2 = segments[:, 2]
   y2 = segments[:, 3]
   b = x1 - x2
   a = y2 - y1
   c = -(a * x1 + b * y1)

   # Ensure that the normal vectors are always
   # pointing toward positive y
   where_b_is_neg = b < 0
   c[where_b_is_neg] *= -1.0
   b[where_b_is_neg] *= -1.0
   a[where_b_is_neg] *= -1.0

   # The length of the segment
   norm = np.sqrt(a**2 + b**2)
   return np.stack([x1, y1, x2, y2, a/norm, b/norm, c/norm, norm], axis=1)


def draw_segments(img, segments, color, thickness):
   '''
       img : cv2 image
       segments : np.array(num_lines, 8)
       color  : 3 element tuple
       thickness : int
   '''
   for s in segments:
       cv2.line(img, (s[0], s[1]), (s[2], s[3]), color, thickness)


def draw_lines(img, segments, color, thickness):
   '''
       img : cv2 image
       segments : np.array(num_lines, 8)
       color  : 3 element tuple
       thickness : int
   '''

   for s in segments:
       a, b, c = s[4:7]
       # If the line is horizontal
       if a == 0:
           cv2.line(img, (0, -int(c/b)), (img.shape[1]-1, -int((c+a*(img.shape[1]-1))/b)), color, thickness)
       # If the line is vertical
       elif b == 0:
           cv2.line(img, (-int(c/a), 0), (-int(c/a), img.shape[0] - 1), color, thickness)
       else:
           # We look for the intersection point with the horizontal axis y=0 and y=height-1
           y = 0
           x = -c/a   #-(c - b*y)/a
           if x < 0:
               x = 0
               y = -(c + a * x) / b
           elif x >= img.shape[1]:
               x = img.shape[1] - 1
               y = -(c + a * x) / b
           first_point = (int(x), int(y))

           y = img.shape[0] - 1
           x = -(c+b*y)/a
           if x < 0:
               x = 0
               y = -(c + a * x) / b
           elif x >= img.shape[1]:
               x = img.shape[1] - 1
               y = -(c + a * x) / b
           second_point = (int(x), int(y))

           cv2.line(img, first_point, second_point, color, thickness)

def intersections(segments):
   intersections = []
   for i, si in enumerate(segments):
       for sj in segments[i+1:]:
           cross_product = np.cross(si[4:6], sj[4:6]) # [a1,b1] ^ [a2, b2]
           if cross_product != 0:
               coeff = 1.0 / cross_product

               intersections.append([coeff * np.cross(si[5:7]   , sj[5:7]), # [b1, c1] ^ [b2, c2]
                                     coeff * np.cross(sj[[4, 6]], si[[4, 6]])]) # -[a1, c1] ^ [a2, c2]
   return np.array(intersections)


def hierarchical_clustering(normal_vectors):
   Z = linkage(normal_vectors, method='ward')  # 'ward' minimizes variance, use other methods if needed
   return fcluster(Z, 2, criterion='maxclust')

def vanishing_lines(lines, normal_vectors):
   # Perform hierarchical/agglomerative clustering
   # Extract two largest clusters
   labels = hierarchical_clustering(normal_vectors)
   lines = np.concatenate((lines[labels == 1], lines[labels == 2]),axis=0)
   normal_vectors = np.concatenate((normal_vectors[labels == 1], normal_vectors[labels == 2]),axis=0)

   kmeans = KMeans(n_clusters=2, random_state=0).fit(normal_vectors)
   labels = kmeans.labels_
   # Séparer les lignes en fonction des clusters
   return lines[labels == 0], lines[labels == 1]

def horizontal_misplacement(frame, vanishing_point):

   ans=vanishing_point[0]/frame.shape[1]-0.5
   if ans<-0.15:
       return -0.25
   if ans>0.15:
       return 0.25
   return 0

def angle_indicator(line1, line2): #Ratio between left and right angles
   agl1 = np.pi/2 - np.arccos(abs(line1[0]))
   agl2 = np.pi/2 - np.arccos(abs(line2[0]))
   # Return the ratio of the two angles
#    return agl1/agl2
   agl=agl1/agl2
   if agl<0.7:
       return -0.5
   if agl>1.5:
       return 0.5
   return 0


import cv2
import numpy as np

def draw_function(
    img,
    Y,                  # Values of the curve, len(Y) = img.shape[1]
    hmin, hmax,         # Curve is drawn between rows hmin and hmax of the image
    ymin, ymax,         # The range [ymin, ymax] is mapped to [hmin, hmax]
    color=(0, 255, 0),  # Line color
    thickness=2         # Line thickness
):
    # Handle invalid values to prevent division by zero or NaN issues
    Y = np.nan_to_num(Y, nan=0.0, posinf=0.0, neginf=0.0)
    if ymin == ymax:
        ymin, ymax = min(Y), max(Y)
        if ymin == ymax:
            ymin, ymax = ymin - 1, ymax + 1  # Ensure a valid range

    # Align the X-axis according to the image width and curve length
    width = img.shape[1]
    x_values = np.linspace(0, width - 1, len(Y)).astype(int)
    
    y_scaled = np.clip(
        hmax - ((Y - ymin) / (ymax - ymin)) * (hmax - hmin), hmin, hmax
    ).astype(int)
    
    for i in range(1, len(x_values)):
        pt1 = (x_values[i - 1], y_scaled[i - 1])
        pt2 = (x_values[i], y_scaled[i])
        if 0 <= pt1[0] < img.shape[1] and 0 <= pt1[1] < img.shape[0] and \
           0 <= pt2[0] < img.shape[1] and 0 <= pt2[1] < img.shape[0]:
            cv2.line(img, pt1, pt2, color, thickness)
    return img



import numpy as np
from scipy.signal import medfilt
from scipy.ndimage import median_filter
from scipy.ndimage import label


def local_shift_computation(curve1, curve2, max_shift=10, window_radius=10, smooth_kernel_size=21, variance_threshold=50):
    cut_curve1 = curve1[max_shift + window_radius : - (max_shift + window_radius)]
   
    x_indices = np.arange(len(cut_curve1))
    s_values = np.arange(max_shift + 1)
    h_values = np.arange(-window_radius, window_radius + 1)

    pos = h_values + max_shift + window_radius
    pos1 = x_indices[:, None] + pos
    pos2 = x_indices[:, None, None] + s_values[None, :, None] + pos

    curve1_window = curve1[pos1]
    curve2_window = curve2[pos2]
   
    diff_squared = (curve1_window[:, None, :] - curve2_window) ** 2

    total_shift = diff_squared.sum(axis=-1)

    optimal_shifts = total_shift.argmin(axis=1)

    padding = max_shift + window_radius
    optimal_shifts = np.pad(optimal_shifts, (padding, padding), mode='constant', constant_values=0)

    smoothed_shifts = medfilt(optimal_shifts, kernel_size=smooth_kernel_size)
    
    return smoothed_shifts


def detect_door(intensity_profile, shift_values, margin, window_length=5):
    """
    Wall (0) or door/depth (1).
    """

    base_ratio = 0.35
    percentile_threshold = 4 * window_length
    segment_width_limit = len(intensity_profile) * 0.30
    filter_window = 30

    assert window_length % 2 == 1
    half_window_length = window_length // 2

    binary_classification = np.zeros_like(shift_values)
    max_shift_value = np.max(shift_values)

    # Generate central indices, avoiding edges
    center_idx = np.arange(margin, len(shift_values) - margin)
    # Create a mask to filter low-shift regions
    low_shift_region = shift_values[center_idx] <= max_shift_value * base_ratio
    target_indices = center_idx[low_shift_region]

    # Use a sliding window to create subsets of the intensity line
    intensity_windows = np.lib.stride_tricks.sliding_window_view(intensity_profile, window_length)
    selected_windows = intensity_windows[target_indices - half_window_length]

    standard_deviations = np.std(selected_windows, axis=1)
    if standard_deviations.size > 0:
        deviation_threshold = np.percentile(standard_deviations, percentile_threshold)
        # Identify points with standard deviation above the threshold
        binary_classification[target_indices] = (standard_deviations > deviation_threshold).astype(int)
        binary_classification = median_filter(binary_classification, size=filter_window)
    else:
        binary_classification.fill(0)

    # Merge adjacent classified segments
    labeled_segments, segment_count = label(binary_classification == 1)
    for segment_label in range(1, segment_count + 1):
        segment_size = np.sum(labeled_segments == segment_label)
        # Remove segments smaller than the width limit
        if segment_size <= segment_width_limit:
            binary_classification[labeled_segments == segment_label] = 0

    return binary_classification

