import rclpy
from rclpy.node import Node
from sensor_msgs.msg import CompressedImage
from nav_msgs.msg import Odometry
import cv2
import numpy as np
from cv_bridge import CvBridge
from vanishing_point.my_line_library import local_shift_computation, draw_function, detect_door

class OpticalFlowNode(Node):
    def __init__(self):
        super().__init__('optical_flow_node')
        
        self.bridge = CvBridge()
        self.current_image = None  
        self.drone_speed = 0.0     
        self.padding = 10          
        self.window_size = 5      

        self.image_sub = self.create_subscription(
            CompressedImage,
            'bebop/image_raw/compressed',
            self.image_callback,
            10
        )
        
        self.odom_sub = self.create_subscription(
            Odometry,
            'bebop/odom',
            self.odom_callback,
            10
        )
        
        self.timer = self.create_timer(0.1, self.process_frame)

    def image_callback(self, msg):
        self.current_image = self.bridge.compressed_imgmsg_to_cv2(msg, "bgr8")

    def odom_callback(self, msg):
        self.drone_speed = np.sqrt(
            msg.twist.twist.linear.x ** 2 +
            msg.twist.twist.linear.y ** 2 +
            msg.twist.twist.linear.z ** 2
        )
        
    def detect_free_space(self, shift, door_detection):
        return np.any(door_detection == 1)

    def process_frame(self):
        if self.current_image is not None:
            
            gray_image = cv2.cvtColor(self.current_image, cv2.COLOR_BGR2GRAY)
            
            curve1 = gray_image[100]
            curve2 = gray_image[101]

            shift = local_shift_computation(curve1, curve2)
            door_detection = detect_door(curve2, shift, self.padding, self.window_size)

            img_with_curves = self.current_image.copy()

            overlap_hmin, overlap_hmax = img_with_curves.shape[0] - 100, img_with_curves.shape[0] - 50
            min_curve = min(np.min(curve1), np.min(curve2))
            max_curve = max(np.max(curve1), np.max(curve2))

            # curve1（yellow）and curve2（green）
            img_with_curves = draw_function(
                img_with_curves, curve1, hmin=overlap_hmin, hmax=overlap_hmax,
                ymin=min_curve, ymax=max_curve, color=(0, 255, 255), thickness=1
            )
            img_with_curves = draw_function(
                img_with_curves, curve2, hmin=overlap_hmin, hmax=overlap_hmax,
                ymin=min_curve, ymax=max_curve, color=(0, 255, 0), thickness=1
            )

            # shift（blue）
            img_with_shift = draw_function(
                img_with_curves, shift, hmin=50, hmax=150,
                ymin=np.min(shift), ymax=np.max(shift), color=(255, 0, 0), thickness=2
            )

            # detect_door（red）
            img_with_door = draw_function(
                img_with_shift, door_detection, hmin=150, hmax=200,
                ymin=0, ymax=1, color=(0, 0, 255), thickness=2
            )

            free_space = self.detect_free_space(shift, door_detection)
            status_text = "Free" if free_space else "Blocked"
            cv2.putText(
                img_with_door, status_text, (50, 50), cv2.FONT_HERSHEY_SIMPLEX,
                1, (0, 255, 0) if free_space else (0, 0, 255), 2
            )

            cv2.imshow("Optical Flow with Curves, Shift, and Door Detection", img_with_door)
            cv2.waitKey(0)

    def run(self):
        rclpy.spin(self)

def main(args=None):
    rclpy.init(args=args)
    node = OpticalFlowNode()
    node.run()
    rclpy.shutdown()

if __name__ == '__main__':
    main()

